<?php
/**
 * zogan - PHP template engine for XML
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * @author Haruki Setoyama <haruki@planewave.org>
 * @copyright Copyright &copy; 2003-2004, Haruki SETOYAMA <haruki@planewave.org>
 * @license http://www.opensource.org/licenses/lgpl-license.php LGPL
 * @version 0.1- $Id: zogan.php,v 1.7 2004/06/17 09:00:39 haruki Exp $
 * @link http://phpoot.sourceforge.jp/zogan/
 * @package zogan
 */
/**
 * Requires PEAR
 */
require_once 'PEAR.php';

// Constants
/**
 * mark for compiled template that mean ZOGAN library has been included
 */
define('ZOGAN', true);
/**
 * directory where ZOGAN library is in
 */
@define('ZOGAN_DIR', dirname(__FILE__).'/zogan');
/**
 * directory where data-formatting library is in
 */
@define('ZOGAN_FORMAT', ZOGAN_DIR.'/format/');
/**
 * extention added to compiled template
 */
@define('ZOGAN_EXT', '.php');
/**
 * parser to compiled template
 */
@define('ZOGAN_PARSER', 'zogan.htmlparser.php');
//@define('ZOGAN_PARSER', 'zogan.parser.php');

/**
 * error code
 */
define('ZOGAN_ERR_INTERNAL', 1);
define('ZOGAN_ERR_TEMPLATE', 2);
define('ZOGAN_ERR_COMPILE_DIR', 4);
define('ZOGAN_WARN_TEMPLATE', 8);

/**
 * ZOGAN Main Class.
 *
 * @access public
 * @package zogan
 * @subpackage interface
 */
class zogan extends PEAR {

    /**
     * options
	 * 
     * @access private
     * @var array
     */
    var $option;

    /**
     * path to directory where compiled files will be stored
     * @var string
     * @access private
     */
    var $_compiled_dir;

    /**
     * path to compiled file
     * @var string
     * @access private
     */
    var $_compiled_file;

    /**
     * path to template file
     * @var zogan_templatefile
     * @access private
     */
    var $_template_file;

    /**
     * variable names not to be escaped
     * @access private
     */
    var $_no_escape;

    /**
     * functions to apply for model data
     * @access private
     */
    var $_functions;

    /**
     * constructor
     */
    function zogan()
    {
        $this->option
            = array(
                'force_compile'   => false,
                'check_timestamp' => true,
                'debug'           => false,
                'mode'            => 0666,
                'flock'           => true,
                'no_escape'       => array(),
                'pi_keep'         => array('xml'),
                'functions'       => array()
            );
    }

    /**
     * sets the path for the template file to display
     *
     * @param stirng $template
     * @return void
     */
    function setTemplateFile($template_file)
    {
        if (is_string($template_file)) {
            $this->_template_file = new zogan_templatefile_local($template_file);
        }
        elseif (is_a($template_file, 'zogan_templatefile')) {
            $this->_template_file = $template_file;
        }
        else {
            return $this->raiseError(
                "Invalid template file. ",
                ZOGAN_ERR_TEMPLATE);
        }

        $this->_set_compiled_file();
        return true;
    }

    /**
     * sets the path for the directory where compiled files will be stored
     *
     * @param stirng $dirname directory name
     * @return true|PEARError
     */
    function setCompileDir($dirname)
    {
        if (! $this->_mkdirs($dirname)) {
            return $this->raiseError(
                "Failed to create the directory for compiled files '$dirname'. ",
                ZOGAN_ERR_COMPILE_DIR);
        }
        $this->_compiled_dir = $dirname;
        $this->_set_compiled_file();
        return true;
    }

    /**
     * makes directory for compiled files 
     *
     * @param stirng $path
     * @access private
     * @return bool
     */
    function _mkdirs($path)
    {
        if (is_dir($path)) return true;
        if (is_file($path)) return false;
        if (! $this->_mkdirs(dirname($path))) return false;
        $ret = mkdir($path);
        if ($ret && isset($this->option['mode'])) {
            chmod($path, $this->option['mode']);
        }
        return $ret;
    }

    /**
     * sets the file name of compiled file
     *
     * @access private
     * @return string
     */
    function _set_compiled_file()
    {
        if (isset($this->_template_file) && isset($this->_compiled_dir)) {
            $this->_compiled_file
                = $this->_compiled_dir . '/'
                . $this->_template_file->getBasename() . ZOGAN_EXT;
        } else {
            $this->_compiled_file = null;
        }
    }

    /**
     * sets option
     * 
	 * <pre>
	 * Key                  Meaning. defalult.
     * 'force_compile'      when true, template will be compiled everytime. false.
     * 'check_timestamp'    when true, modified time of template will be checked. true.
     * 'debug'              when true, debug information displayed and compile file removed. false.
     * 'mode'               mode for compiled files and directories. 0777.
     * </pre>
     * @param string $key    key for option
     * @param mixed $val    value of option
     * @return void
     */
    function setOption($key, $val)
    {
        if (isset($this->option[$key])) {
            $this->option[$key] = $val;
            return true;
        } else {
            return false;
        }
    }

    /**
     * sets the variable names not to be escaped
     *
     * @param array 
     * @return string
     */	
    function setNoEscapeVars($vars)
    {
        $this->_no_escape = array();
        foreach ($vars as $var) {
            $this->_no_escape[$var] = 1;
        }
    }

    /**
     * sets the function resouces to modify model data
     *
     * @param array 
     * @return string
     */	
    function setFunctions($funcs)
    {
        $this->_functions = $funcs;
    }

    /**
     * displays the template results
     *
     * @param array $data   model data
     * @return bool|PEARError
     */
    function display($data)
    {
        if ($this->_need_compile() !== false) {
            // compile
            $err = $this->compile();
            if ($err !== true) {
                return $err;
            }
        }

        if ($this->option['debug']
        || $this->_default_error_mode & PEAR_ERROR_PRINT) {
            $this->_quiet = false;
        } else {
            $this->_quiet = true;
        }

        $err = $this->_display_compiled($data);
        if ($err !== true) {
            $this->removeCompiled();
            return $this->raiseError(
                    "Failed to run the compiled file '$err'. ",
                    ZOGAN_ERR_INTERNAL);
        }

        if ($this->option['debug'] != false) {
            $this->removeCompiled();
        }

        return true;
    }

    /**
     * displays the template results with compiled template
     *
     * @param array     $_val0          model data
     * @access private
     * @return bool
     */
    function _display_compiled($_val0)
    {
        $_check = $this->_compiled_file; // flag for parse error check
        include $this->_compiled_file;
        return $_check;
    }

    /**
     * checks if the template needs to be compiled
     *
     * @access private
     * @return bool
     */
    function _need_compile()
    {
        if ($this->option['force_compile'] != false
        || $this->option['debug'] != false) {
            return true;
        }
        if (file_exists($this->_compiled_file) != true) {
            return true;
        }
        if ($this->option['check_timestamp'] != false
        && $this->_template_file->getUpdateTime() > filemtime($this->_compiled_file)) {
            return true;
        }
        return false;
    }

    /**
     * gets the template results as string
     *
     * @param sting $output     variable for output
     * @param array $data       model data
     * @return bool|PEARError
     */
    function expand(&$output, $data)
    {
        ob_start();
        $ret = $this->display($data);
        $output = ob_get_contents();
        ob_end_clean();
        return $ret;
    }

    /**
     * compiles the template file
     *
     * @return bool
     */
    function compile()
    {
        require_once ZOGAN_DIR.'/localfile.php';

        $file = new LocalFile($this->_compiled_file);
        $file->flock = $this->option['flock'];

        $lock = $file->lock();
        if ($lock !== true) { return $lock; }

        $parsed = $this->_parse();
        if (PEAR::isError($parsed)) {
            $file->unlock();
            return $parsed;
        }
        $parsed = "<?php if(! defined('ZOGAN')) exit(); ?>\n"
                    . $parsed . "<?php \$_check = true; ?>\n"; // <?

        return $file->write($parsed, $this->option['mode']);
    }

    /**
     * removes the compiled template file
     *
     * @return bool
     */
    function removeCompiled()
    {
        if (! isset($this->_template_file)) {
            return $this->raiseError(
                    "Template file is not set. ",
                    ZOGAN_ERR_TEMPLATE);
        }

        $compiled_file = $this->_compiled_file;
        if (! file_exists($compiled_file)) return false;

        if (! unlink($compiled_file)) {
            return $this->raiseError(
                    "Failed to remove the compiled file '$compiled_file'. ",
                    ZOGAN_ERR_COMPILE_DIR);
        }
        return true;
    }

    /**
     * parses the template file
     *
     * @access private
     * @return string|PEARError parsed PHP code
     */
    function _parse()
    {
        $template = $this->_template_file->getTemplate();
        if (PEAR::isError($template)) return $template;

        require_once ZOGAN_DIR . '/' . ZOGAN_PARSER;
        $parser = new zogan_parser($this, $this->_template_file->getEncoding());

        return $parser->build($template, $this->option);
    }

    /**
     * sets option for content formating
     */
    function setFormatOption($type, $option)
    {
        require_once ZOGAN_FORMAT.$type.'.php';
        if (! function_exists('zogan_format_'.$type.'_option')) {
            return $this->raiseError(
                "Failed to set format options for $type. ",
                ZOGAN_ERR_INTERNAL);
        }
        return call_user_func('zogan_format_'.$type.'_option', $option);
    }

    /**
     * escapes string with XML predefined entity
     * @return string
     * @access private
     */
    function _escape($val)
    {
        return htmlspecialchars($val);
        /*return str_replace(array('<','>','&','\'','"')
                            , array('&lt;', '&gt;', '&amp;', '&apos;', '&quot;')
                            ,$val);*/
    }

    // functions below used in compiled template file.

    /**
     * returns array of data for nested variabe tag
     * not hash!
     * @return array
     * @access private
     */
    function _c_each($val, $var)
    {
        $target = $this->_c_val($val, $var);
        if (is_array($target) && isset($target[0])) {
            return $target;
        }
        return array($target);
    }

    /**
     * change scalar and array data to hash
     * @access private
     */
    function _c_hash(&$scalar, $var)
    {
        if (is_scalar($scalar)) {
            $arr[$var] = $scalar;
            $scalar = $arr;
        }
        elseif (is_array($scalar) && isset($scalar[0])) {
            $temp = array();
            foreach($scalar as $val)
            {
                $temp[] = array($var => $val);
            }
            $scalar = $temp;
        }
    }

    /**
     * check if $val is printable, that is string or numeric
     * @access private
     */
    function _c_is_string($val)
    {
        return (is_string($val) || is_numeric($val));
    }

    /**
     * print waring for invalid data type
     * @access private
     */
    function _c_warn()
    {
        echo (! $this->_quiet) ? '<!-- INVALID DATA -->' : '';
    }

    /**
     * makes string from data for attribute
     * @return string
     * @access private
     */
    function _c_attr($attr, $target, $default)
    {
        if ($target === true) {
            return ' ' . $attr . '="' . htmlspecialchars($default) . '"';
        } elseif ($this->_c_is_string($target)) {
            return ' ' . $attr . '="' . htmlspecialchars($target) . '"';
        } else {
            return '';
        }
    }

    /**
     * get value from $val with variable name
     * @return mixed
     * @access private
     */
    function _c_val($val, $var)
    {
        if (is_array($val)) {
            if (isset($val[$var])) {
                return $val[$var];
            }
        }
        if (is_object($val)) {
            if (is_callable(array($val, $var))) {
                return $val->$var();
            } else {
                return $val->$var;
            }
        }
        return null;
    }

    /**
     * print comment for no data
     * @access private
     */
    function _c_nodata($indent, $linefeed)
    {
        echo (! $this->_quiet) ? $indent . '<!-- no data -->' . $linefeed : '';
    }

    /**
     * escape XML predefined entity
     * @return string
     * @access private
     */
    function _c_escape($val, $variable)
    {
        // TODO: unescape when needed
        if (isset($this->_no_escape[$variable])) {
            return $val;
        } else {
            return $this->_escape($val);
        }
    }

    /**
     * change basename of template for file inclution
     * @return bool
     * @access private
     */
    function _c_change_basename($basename)
    {
        $this->_template_file->changeBasename($basename);
        $this->_set_compiled_file();
        return true;
    }

    /**
     * apply function to model data
     * @return mixed
     * @access private
     */
    function _c_func($val, $func_name, $arrr =false)
    {
        if (isset($this->_functions[$func_name])) {
            return call_user_func($this->_functions[$func_name], $val);
        }
        elseif ($arrr != false || !$this->_c_is_string($val)) {
            return $val;
        }
        else {
            return $this->_escape($val);
        }
    }
}

/**
 * ZOGAN template file abstraction
 *
 * @access public
 * @package zogan
 * @subpackage interface
 * @abstract
 */
class zogan_templatefile extends PEAR {

    /**
     * constructor
     */
    function zogan_templatefile($file) {}

    /**
     * returns update time of template
     * @return int UNIX time
     */
    function getUpdateTime() {}

    /**
     * returns basename of template file
     * This name will be used for compiled file
     * @return string
     */
    function getBasename() {}

    /**
     * returns template string
     * @return string
     */
    function getTemplate() {}

    /**
     * change only basename of template
     * @return void
     */
    function changeBasename($file) {}

    /**
     * returns charactor encoding of template.
     * UTF-8 is recommended for XML template.
     * @return string
     */
    function getEncoding() {}
}

/**
 * ZOGAN template file abstraction for local file
 *
 * @access private
 * @package zogan
 * @subpackage interface
 */
class zogan_templatefile_local extends zogan_templatefile {

    var $_file;
    var $_file_get_contents;

    function zogan_templatefile_local($file)
    {
        $this->_file = $file;
        $this->_file_get_contents = function_exists('file_get_contents');
    }

    function getUpdateTime()
    {
        return filemtime($this->_file);
    }

    function getBasename()
    {
        return basename($this->_file);
    }

    function getTemplate()
    {
        if ($this->_file_get_contents) {
            $content = file_get_contents($this->_file);
        } else {
            $fp = fopen($this->_file, 'rb');
            if ($fp === false) {
                return $this->raiseError(
                    "Failed to open a file '".$this->_file."' for reading. ",
                    ZOGAN_ERR_TEMPLATE);
            }
            $content = fread($fp, filesize($this->_file));
        }
        if ($content === false) {
            return $this->raiseError(
                "Failed to get content from a file '".$this->_file."'. ",
                ZOGAN_ERR_TEMPLATE);
        }
        return $content;
    }

    function changeBasename($file)
    {
        $this->_file = dirname($this->_file).'/'.$file;
    }

    function getEncoding()
    {
        return 'utf-8';
    }
}

/**
 * function to clone object
 */
if ((int)substr(PHP_VERSION,0,1) >= 5) {
    /**
     * @access private
     */
    function zogan_clone_obj($obj)
    {
        return clone($obj);
    }
} else {
    /**
     * @access private
     */
    function zogan_clone_obj($obj)
    {
        return $obj;
    }
}

?>